use std::{collections::HashMap, fmt::Display};

use crate::{
    callable::{Callable, CallingError},
    function::Function,
    instance::Instance,
    value::Value,
};

#[derive(Debug, Clone, PartialEq)]
pub struct Class {
    pub name: String,
    pub methods: HashMap<String, Function>,
}

impl Display for Class {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "{}", self.name)
    }
}

impl Class {
    pub fn new(name: String, methods: HashMap<String, Function>) -> Self {
        Self { name, methods }
    }

    pub fn find_method(&self, name: &str) -> Option<&Function> {
        self.methods.get(name)
    }
}

impl Callable for Class {
    fn name(&self) -> String {
        self.name.clone()
    }

    fn arity(&self) -> usize {
        if let Some(initializer) = self.find_method("init") {
            initializer.arity()
        } else {
            0
        }
    }

    fn call(
        &self,
        interpreter: &mut crate::interpreter::Interpreter,
        args: Vec<crate::value::Value>,
    ) -> Result<Value, CallingError> {
        let class = self.clone();
        let instance = Instance::new(class);
        if let Some(initializer) = self.find_method("init") {
            initializer.bind(instance.clone()).call(interpreter, args)?;
        };

        Ok(Value::Instance(instance))
    }
}
